/*setup includes*/
#include <windows.h>
#include <stdio.h>
#include <commctrl.h>
#include "tasks.h"
#include "../ls-b24/lsapi/lsapi.h"

/*setup constants*/
const char AppName[] = "Tasks";
const int MAXTASKS = 16;
const int MAXICONS = 64;

/*setup variables*/
IconList iconTag[MAXICONS];
IgnoreList ignoreTag[MAXTASKS];
PixList pixTag[MAXICONS];
TasksSettings ts;
TaskType tasks[MAXTASKS];

BOOL taskWrap, bgImgNone, selBgImgNone;
BOOL moveOnce, mouseActive;
char path[MAXLEN];
HCURSOR normalCursor, movingCursor;
HINSTANCE dll;
HWND parent, mainWin, deskTop, hints, curTask, lockCurTask, oldCurTask;
int numIcon, numIgnore, numPix, numTasks, prevTasks;
int screenHeight, screenWidth;
POINT oldCursorPos;
UINT timer;

/*declare all the functions*/
BOOL GetTasksIgnore(char *ignoreTitle);
BOOL IsAppWindow(HWND hWNd);
HICON GetIconFromWindow(HWND hWnD, BOOL bBigIcon);
LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);
int GetTaskByWnd(HWND Hwnd);
int GetUnMovedNrTasks(void);
void CreateHints(HWND hWnd, char *txt, RECT *r);
void CreateTask(int i, HWND handle);
void DeleteTask(int i);
void GetNewTaskPos(int i);
void GetTasks(void);
void GetTasksIcon(int n);
void GetTasksPix(int n);
void PaintTask(int i);
void ReadConfig(void);
void ReadTasksIcon(void);
void ReadTasksIgnore(void);
void ReadTasksPix(void);
void ReArrangeTasks(BOOL all);
void RemoveHints(HWND hWnd);
void TaskPosition(int i, int tmpNumTasks);
/*The rest are !bang commands*/
void BottomTasks(void);
void GatherTasks(void);
void HideTasks(void);
void ShowTasks(void);
void ToggleTasks(void);
void TopTasks(void);

/*and here are all the functions*/
int initModule(HWND ParentWnd, HINSTANCE dllInst, wharfDataType* wd)
{
  return initModuleEx (ParentWnd, dllInst, wd->lsPath);
}

TasksSettings ReadSettings()
{
  TasksSettings settings;
  char tmp[MAXLEN], tmp2[MAXLEN], tmp3[MAXLEN], tmp4[MAXLEN];

  GetRCString("TasksDirection", tmp, "right", MAXLEN);
	  settings.Direction = 2;  /*only needed if user enters an invalid parameter*/
	if (strcmp(tmp, "up") == 0)
	  settings.Direction = 1;
	else if (strcmp(tmp, "right") == 0)
	  settings.Direction = 2;
	else if (strcmp(tmp, "down") == 0)
	  settings.Direction = 3;
	else if (strcmp(tmp, "left") == 0)
	  settings.Direction = 4;
  settings.DragDistance = GetRCInt("TasksDragDistance", 4);
  settings.PaddingX = GetRCInt("TasksPaddingX", 0);
  settings.PaddingY = GetRCInt("TasksPaddingY", 0);
  settings.WrapCount = GetRCInt("TasksWrapCount", MAXTASKS + 1);
	if (settings.WrapCount > MAXTASKS || settings.WrapCount < 2)
	{
	  settings.WrapCount = MAXTASKS;
	  taskWrap = false;
	}
	else
	  taskWrap = true;
  GetRCString("TasksWrapDirection", tmp2, "down", MAXLEN);
	  settings.WrapDirection = 3;  /*only needed if user enters an invalid parameter*/
	if (strcmp(tmp2, "up") == 0)
	  settings.WrapDirection = 1;
	else if (strcmp(tmp2, "right") == 0)
	  settings.WrapDirection = 2;
	else if (strcmp(tmp2, "down") == 0)
	  settings.WrapDirection = 3;
	else if (strcmp(tmp2, "left") == 0)
	  settings.WrapDirection = 4;

  settings.Height = GetRCInt("TasksHeight", 32);
    if (settings.Height < 2) settings.Height = 32;
  settings.Width = GetRCInt("TasksWidth", 32);
    if (settings.Width < 2) settings.Width = 32;
  settings.X = GetRCInt("TasksX", 0);
  settings.Y = GetRCInt("TasksY", 0);

  settings.MinimizedX = GetRCInt("TasksMinimizedX", 0);
  settings.MinimizedY = GetRCInt("TasksMinimizedY", 0);

  settings.IconSize = GetRCInt("TasksIconSize", 16);
    if (settings.IconSize < 2) settings.IconSize = 16;
  settings.IconX = GetRCInt("TasksIconX", -2);
	if (settings.IconX < 0)
	{
	  if (settings.IconX == -1)
		settings.IconX = 1;
	  else if (settings.IconX == -3)
		settings.IconX = settings.Width - settings.IconSize - 1;
	  else /*Center the Icon for '-2' and by default for any other negative value.*/
		settings.IconX = settings.Width/2 - settings.IconSize/2;
	}
  settings.IconY = GetRCInt("TasksIconY", -2);
	if (settings.IconY < 0)
	{
	  if (settings.IconY == -1)
		settings.IconY = 1;
	  else if (settings.IconY == -3)
		settings.IconY = settings.Height - settings.IconSize - 1;
	  else /*Center the Icon for '-2' and by default for any other negative value.*/
		settings.IconY = settings.Height/2 - settings.IconSize/2;
	}

  settings.PixX = GetRCInt("TasksPixX", 1);
  settings.PixY = GetRCInt("TasksPixY", 1);
		
  settings.AlwaysOnTop = GetRCBool("TasksAlwaysOnTop", TRUE);
  settings.AutoArrange = GetRCBool("TasksAutoArrange", TRUE);
  settings.GroupMinimized = GetRCBool("TasksGroupMinimized", TRUE);
  settings.OnlyMinimized = GetRCBool("TasksOnlyMinimized", TRUE);
  settings.MoveAll = GetRCBool("TasksMoveAll", TRUE);
  settings.NoMove = GetRCBool("TasksNoMove", TRUE);
  settings.StartHidden = GetRCBool("TasksStartHidden", TRUE);

  settings.NoHints = GetRCBool("TasksNoHints", TRUE);
  settings.HighLightSelected = GetRCBool("TasksHighLightSelected", TRUE);
  settings.ShadeMinimized = GetRCBool("TasksShadeMinimized", TRUE);

  settings.BgColor = GetRCColor("TasksBgColor", 0xFF00FF);
  settings.DarkColor = GetRCColor("TasksDarkColor", 0xFF00FF);
  settings.LightColor = GetRCColor("TasksLightColor", 0xFF00FF);
  settings.MinShadeColor = GetRCColor("TasksMinShadeColor", 0x0000FF);
  settings.SelBgColor = GetRCColor("TasksSelBgColor", settings.BgColor);
  settings.SelDarkColor = GetRCColor("TasksSelDarkColor", settings.DarkColor);
  settings.SelLightColor = GetRCColor("TasksSelLightColor", settings.LightColor);

  GetRCString("TasksBackImage", tmp3, ".none", MAXLEN);
	bgImgNone = true;
	if (strcmp(".none", tmp3))
	{
	  settings.BackImage = LoadLSImage(tmp3, NULL);
	  bgImgNone = false;
	}
	else
	  settings.BackImage = NULL;

  GetRCString("TasksSelBackImage", tmp4, "", MAXLEN); /*By using "" as the default, it uses the TasksBackImage.*/
	selBgImgNone = false;
	if (strcmp(".none", tmp4))
	  settings.SelBackImage = LoadLSImage(tmp4, NULL);
	else
	{
	  settings.SelBackImage = NULL;
	  selBgImgNone = true;
	}

  settings.TitleIconX = GetRCInt("TasksTitleIconX", -2);
	if (settings.TitleIconX < 0)
	{
	  if (settings.TitleIconX == -1)
		settings.TitleIconX = 1;
	  else if (settings.TitleIconX == -3)
		settings.TitleIconX = settings.Width - settings.IconSize - 1;
	  else /*Center the Icon for '-2' and by default for any other negative value.*/
		settings.TitleIconX = settings.Width/2 - settings.IconSize/2;
	}
  settings.TitleIconY = GetRCInt("TasksTitleIconY", -1);
	if (settings.TitleIconY < 0)
	{
	  if (settings.TitleIconY == -1)
		settings.TitleIconY = 1;
	  else if (settings.TitleIconY == -3)
		settings.TitleIconY = settings.Height - settings.IconSize - 1;
	  else /*Center the Icon for '-2' and by default for any other negative value.*/
		settings.TitleIconY = settings.Height/2 - settings.IconSize/2; //Center the Icon by default for any other negative value.
	}
  settings.TitlePixX = GetRCInt("TasksTitlePixX", 1);
  settings.TitlePixY = GetRCInt("TasksTitlePixY", 1);

  settings.TitleHeight = GetRCInt("TasksTitleHeight", 12); /*The Title settings here align it centered along the bottom of a 32x32 tile.*/
    if (settings.TitleHeight < 2) settings.TitleHeight = 12;
  settings.TitleWidth = GetRCInt("TasksTitleWidth", 28);
    if (settings.TitleWidth < 2) settings.TitleWidth = 28;
  settings.TitleX = GetRCInt("TasksTitleX", 2);
  settings.TitleY = GetRCInt("TasksTitleY", 18);

  settings.TitleFontSize = GetRCInt("TasksTitleFontSize", 10);
    if (settings.TitleFontSize < 2) settings.TitleFontSize = 10;

  settings.TitleAlways = GetRCBool("TasksTitleAlways", TRUE);
  settings.TitleOnlyMinimized = GetRCBool("TasksTitleOnlyMinimized", TRUE);

  GetRCString("TasksTitleFont", settings.TitleFont, "Arial", MAXLEN);

  settings.TitleBgColor = GetRCColor("TasksTitleBgColor", 0xFF00FF);
  settings.TitleDarkColor = GetRCColor("TasksTitleDarkColor", 0xFF00FF);
  settings.TitleFontColor = GetRCColor("TasksTitleFontColor", 0xC0C0C0); /*'C0C0C0' is a light gray*/
  settings.TitleLightColor = GetRCColor("TasksTitleLightColor", 0xFF00FF);
  settings.TitleSelBgColor = GetRCColor("TasksTitleSelBgColor", settings.TitleBgColor);
  settings.TitleSelDarkColor = GetRCColor("TasksTitleSelDarkColor", settings.TitleDarkColor);
  settings.TitleSelFontColor = GetRCColor("TasksTitleSelFontColor", settings.TitleFontColor);
  settings.TitleSelLightColor = GetRCColor("TasksTitleSelLightColor", settings.TitleLightColor);
  return settings;
}

int initModuleEx(HWND ParentWnd, HINSTANCE dllInst, LPCSTR szPath)
{
  int i;
  RECT r;
  dll = dllInst;
  strcpy(path, szPath);
  parent = ParentWnd;
  GetClientRect(GetDesktopWindow(),&r);
  screenWidth = r.right;
  screenHeight = r.bottom;
  moveOnce = false;
  mouseActive = false;
  curTask = NULL;
  lockCurTask = NULL;
  oldCurTask = NULL;
  deskTop = NULL;
  deskTop = FindWindow("DesktopBackgroundClass", NULL);
  if (!deskTop)
    deskTop = GetDesktopWindow();
  {
	WNDCLASS wc;
    memset(&wc,0,sizeof(wc));
	wc.lpfnWndProc = WndProc;
	wc.hInstance = dllInst;
	wc.lpszClassName = AppName;
	wc.style = CS_DBLCLKS;
	if (!RegisterClass(&wc)) 
	{
	  MessageBox(parent,"Error registering window class","Tasks",MB_OK);
	  return 1;
	}
  }
  normalCursor = GetCursor();
  if (!normalCursor)
	LoadCursor(NULL, IDC_ARROW);
  movingCursor = LoadCursor(NULL, IDC_SIZEALL);
  for (i = 0; i < MAXTASKS; i++)
  {
	tasks[i].hwnd = NULL;
	tasks[i].pix = NULL;
	tasks[i].appHandle = NULL;
	tasks[i].appTitle = NULL;
	tasks[i].appClass = NULL;
	tasks[i].icon = NULL;
  }
  ts = ReadSettings();
  ReadTasksIcon();
  ReadTasksIgnore();
  ReadTasksPix();
  mainWin = CreateWindowEx(WS_EX_TOOLWINDOW, AppName, "Timer", WS_CHILD, 0,0,0,0, parent, NULL, dll, NULL);
  if (!mainWin) 
  {						   
    MessageBox(parent,"Error creating window","Tasks",MB_OK);
    return 1;
  }
  if (!ts.NoHints)
  {
	hints = CreateWindow(TOOLTIPS_CLASS,(LPSTR) NULL,TTS_ALWAYSTIP,	CW_USEDEFAULT,	CW_USEDEFAULT,CW_USEDEFAULT,CW_USEDEFAULT,NULL,(HMENU) NULL,dll,NULL);
	if (!hints) 
	{
	  MessageBox(parent, "Error creating hints window", "Tasks", MB_OK);
	  return 1;
	}
	SetWindowPos(hints, HWND_TOPMOST,0,0,0,0,SWP_NOMOVE|SWP_NOSIZE|SWP_NOACTIVATE);
  }
  SetWindowLong(mainWin, GWL_USERDATA, magicDWord);
  ShowWindow(mainWin, SW_HIDE);
  AddBangCommand("!BOTTOMTASKS", BottomTasks);
  AddBangCommand("!GATHERTASKS", GatherTasks);
  AddBangCommand("!HIDETASKS", HideTasks);
  AddBangCommand("!SHOWTASKS", ShowTasks);
  AddBangCommand("!TOGGLETASKS", ToggleTasks);
  AddBangCommand("!TOPTASKS", TopTasks);
  timer = SetTimer(mainWin, 0, 250, NULL);
  return 0;
}

void quitModule(HINSTANCE dllInst)
{
  int i;
  if (timer)
    KillTimer(NULL, timer);
  for(i = 0; i < numTasks; i++)
    DeleteTask(i);
  for(i = 0; i < numPix; i++)
  {
	delete [] pixTag[numPix].match;
	delete [] pixTag[numPix].pix;
  }
  for(i = 0; i < numIcon; i++)
  {
	delete [] iconTag[numIcon].match;
	delete [] iconTag[numIcon].name;
  }
  for(i = 0; i < numIgnore; i++)
	delete [] ignoreTag[numIgnore].match;
  if (ts.BackImage)
    DeleteObject(ts.BackImage);
  if (ts.SelBackImage)
	DeleteObject(ts.SelBackImage);
  DestroyWindow(mainWin);
  if (!ts.NoHints) 
	DestroyWindow (hints);
  UnregisterClass(AppName,dllInst);
  return;
}

BOOL GetTasksIgnore(char *ignoreStr)
{
  int i;
  for (i = 0; i < numIgnore; i++)
  {
	if (strstr(ignoreStr, ignoreTag[i].match))
	  return true;
  }
  return false;
}

BOOL IsAppWindow(HWND hWNd)
{
	BOOL bResult = TRUE;
	LONG nStyle;
	HWND hOwner;
	HWND hParent;

	/*if it is a WS_CHILD or not WS_VISIBLE, fail it*/
	nStyle = GetWindowLong(hWNd, GWL_STYLE);	
	if((nStyle & WS_CHILD) || !(nStyle & WS_VISIBLE))
		bResult = FALSE;

	/*if the window is a WS_EX_TOOLWINDOW fail it*/
	nStyle = GetWindowLong(hWNd, GWL_EXSTYLE);
	if(nStyle & WS_EX_TOOLWINDOW)
		bResult = FALSE;

	/*If this has an owner, then only accept this window
	  if the partent is not accepted*/
	hOwner = GetWindow(hWNd, GW_OWNER);
	if(hOwner && bResult)
		bResult = !IsAppWindow(hOwner);

	/*if it has a parent, fail*/
	hParent = GetParent(hWNd);
	if(hParent)
		bResult = FALSE;

	/*if the magicDWord, fail*/
	if(GetWindowLong(hWNd, GWL_USERDATA) == 0x49474541)
		bResult = FALSE;

	return bResult;
}

HICON GetIconFromWindow(HWND hWnD, BOOL bBigIcon) /*taken from the re5ource popups, much credit to re5ource :)*/
{
  HICON a;

  if (!bBigIcon)
  {/*thanks again to fahim's lstasks src*/
	SendMessageTimeout(hWnD, WM_GETICON, 0, 0, 0, 1000, (unsigned long*)&a);
	if (!a) a = (HICON)GetClassLong(hWnD, GCL_HICONSM);
	if (!a) SendMessageTimeout(hWnD, WM_GETICON, 1, 0, 0, 1000, (unsigned long*)&a);
	if (!a) a = (HICON)GetClassLong(hWnD, GCL_HICON);
	if (!a) SendMessageTimeout(hWnD, WM_QUERYDRAGICON, 0, 0, 0, 1000, (unsigned long*)&a);
  } else {
	SendMessageTimeout(hWnD, WM_GETICON, 1, 0, 0, 1000, (unsigned long*)&a);
	if (!a) a = (HICON)GetClassLong(hWnD, GCL_HICON);
	if (!a) SendMessageTimeout(hWnD, WM_QUERYDRAGICON, 0, 0, 0, 1000, (unsigned long*)&a);
	if (!a) SendMessageTimeout(hWnD, WM_GETICON, 0, 0, 0, 1000, (unsigned long*)&a);
	if (!a) a = (HICON)GetClassLong(hWnD, GCL_HICONSM);
  }
  if (!a) a = LoadIcon(NULL, MAKEINTRESOURCE(IDI_APPLICATION));
  return a;
}

LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
  switch (message)
  {
	case WM_TIMER:
	{
	  oldCurTask = curTask;
	  curTask = GetForegroundWindow();
	  if (oldCurTask != curTask)
		lockCurTask = oldCurTask;
	  GetTasks();
      return 0;
	}
//	case WM_QUERYENDSESSION:
//	{
//	  return SendMessage(parent,message,wParam,lParam);
//	}
//	case WM_CREATE:
//	  return 0;
	case WM_ERASEBKGND:
	{
	  if (ts.BgColor == 0x00FF00FF || ts.AlwaysOnTop) /* no point in doing this if not transparent.  But because of a wierd quirk must have it when always on top, -jugg*/
		PaintDesktop((HDC)wParam); // transparency (Maduin)
	  return 1;
	}
	case WM_PAINT:
	{
      int n = GetTaskByWnd(hwnd);
      if (n != -1)
		PaintTask(n);
	  return 0;
	}
	case WM_LBUTTONDOWN:
	{
	  int n = GetTaskByWnd(hwnd);
	  if (n != -1)
	  {
	    GetCursorPos(&oldCursorPos);
	    SetCapture(tasks[n].hwnd);
		moveOnce = true;
		mouseActive = true;
	  }
	  return 0;
	}
	case WM_MOUSEMOVE:
	{
	  int n = GetTaskByWnd(hwnd);
	  if (n != -1)
	  {
		if (mouseActive && !ts.NoMove)
		{
		  POINT cursorPos;
		  GetCursorPos(&cursorPos);
		  if (cursorPos.x == oldCursorPos.x && cursorPos.y == oldCursorPos.y)
		  {
			if (GetCursor() != normalCursor)
			  SetCursor(normalCursor);
		  }
		  else
		  {
			if (GetCursor() != movingCursor)
			  SetCursor(movingCursor);
		  }
		  SetWindowPos(tasks[n].hwnd, 0, tasks[n].x - (oldCursorPos.x - cursorPos.x), tasks[n].y - (oldCursorPos.y - cursorPos.y), ts.Width, ts.Height, SWP_NOZORDER|SWP_NOACTIVATE);
		  if (ts.BgColor == 0x00FF00FF || moveOnce) /*no point in doing this if not transparent, and do it the first move no matter what, so it clears and junk. -jugg*/
			InvalidateRect(hwnd, NULL, TRUE);/*transparency during move, causes some flicker, but it works (Maduin)*/
		  moveOnce = false;
		}
		else
		{
		  if (GetCursor() != normalCursor)
			SetCursor(normalCursor);
		  if (!ts.NoHints) 
		  {
			MSG TooltipMessage;
			TooltipMessage.hwnd=tasks[n].hwnd;
			TooltipMessage.message=message;
			TooltipMessage.wParam=wParam;
			TooltipMessage.lParam=lParam;
			SendMessage(hints,TTM_RELAYEVENT,0,(LPARAM)&TooltipMessage);
			SetWindowPos(hints,HWND_TOPMOST,0,0,0,0,SWP_NOMOVE|SWP_NOSIZE|SWP_NOACTIVATE);
		  }
		}
	  }
	  return 0;
	}
	case WM_LBUTTONUP:
	{
	  POINT cursorPos;
	  GetCursorPos(&cursorPos);
	  ReleaseCapture();
	  int n = GetTaskByWnd(hwnd);
	  if (n != -1)
	  {
		if ((abs(oldCursorPos.x - cursorPos.x) <= ts.DragDistance && abs(oldCursorPos.y - cursorPos.y) <= ts.DragDistance) || ts.NoMove)
		{
		  if (IsIconic(tasks[n].appHandle))
		  {
			ShowWindow(tasks[n].appHandle, SW_RESTORE);
			SetForegroundWindow(tasks[n].appHandle);
			tasks[n].IsActive = true;
		  }
		  else if (tasks[n].appHandle == curTask || (curTask != tasks[n].appHandle && curTask == tasks[n].hwnd && lockCurTask == tasks[n].appHandle))
		  {
			ShowWindow(tasks[n].appHandle, SW_MINIMIZE);
			tasks[n].IsActive = false;
		  }
		  if (!IsIconic(tasks[n].appHandle))
		  {
			SendMessage(parent, LM_BRINGTOFRONT, 0, (long)tasks[n].appHandle);
			SetForegroundWindow(tasks[n].appHandle);
			tasks[n].IsActive = true;
		  }
		  SetWindowPos(tasks[n].hwnd, 0, tasks[n].x, tasks[n].y, ts.Width, ts.Height, SWP_NOZORDER|SWP_NOACTIVATE);
		}
		else if (!ts.NoMove) /*move*/
		{
		  tasks[n].x -= oldCursorPos.x - cursorPos.x;
		  tasks[n].y -= oldCursorPos.y - cursorPos.y;
		  if (ts.MoveAll)
		  {
			if (ts.Direction == 1 || (taskWrap && ts.WrapDirection == 1))
			  ts.Y = tasks[n].y + ts.Height;
			else
			  ts.Y = tasks[n].y;
			if (ts.Direction == 4 || (taskWrap && ts.WrapDirection == 4))
			  ts.X = tasks[n].x + ts.Width;
			else
			  ts.X = tasks[n].x;
			ReArrangeTasks(true);
		  }
		  else
		  {
			SetWindowPos(tasks[n].hwnd, 0, tasks[n].x, tasks[n].y, ts.Width, ts.Height, SWP_NOZORDER|SWP_NOACTIVATE);
			if (ts.AutoArrange && !tasks[n].IsMoved)
			{
			  tasks[n].IsMoved = true;
			  ReArrangeTasks(false);
			}
		  }
		}/*Have to force a complete redraw here...*/
		ShowWindow(tasks[n].hwnd, SW_HIDE);
		ShowWindow(tasks[n].hwnd, SW_SHOWNORMAL);
	  }
	  if (GetCursor() != normalCursor)
		SetCursor(normalCursor);
	  mouseActive = false;
	  lockCurTask = NULL;
	  return 0;
	}
	case WM_MOUSEACTIVATE:
	{
//	  PostMessage(parent, 9185, 0, 0); /*Tells litestep it has been selected*/
//	  SendMessage(parent, 9183, (int)HIWORD(lParam), (int)LOWORD(lParam)); /*Close any popup open*/
	  return MA_ACTIVATE;
	}
	case WM_ACTIVATE:
	{
	  if (LOWORD(wParam) != WA_INACTIVE && (BOOL)HIWORD(wParam) != 0)
		SetActiveWindow(HWND(lParam));
//	  SendMessage(parent, 9185, 0, 0); /*Tells litestep it has been selected*/
//	  if (LOWORD(wParam)) SetActiveWindow(parent);
//	  return 0;
	}
	case WM_RBUTTONUP:
	  return 0;
	case WM_RBUTTONDOWN: /*Open system popup menu - Thanks to Fahims lsTasks src! :)*/
	{
	  BOOL result;
	  int n = GetTaskByWnd(hwnd);
	  if (n != -1)
	  {
		result = TrackPopupMenu(GetSystemMenu(tasks[n].appHandle, FALSE), TPM_RETURNCMD|TPM_RIGHTBUTTON, (int)LOWORD(GetMessagePos()), (int)HIWORD(GetMessagePos()), 0, hwnd, NULL);
		if (result)
		{
		  if (result != SC_MINIMIZE && result != SC_ICON)
			SetForegroundWindow(tasks[n].appHandle);
		  SendMessage(tasks[n].appHandle, WM_SYSCOMMAND, (WPARAM)result, GetMessagePos());
		}
	  }
	  return 0;
	}
	case WM_MBUTTONUP: /*LiteStep poup menu*/
	{
	  SendMessage(parent, 9182, (int)HIWORD(lParam), (int)LOWORD(lParam));
	  return 0;
	}
//	case WM_KEYDOWN: 
//	case WM_KEYUP:
//	  PostMessage(parent,message,wParam,lParam);
//	return 0;
  }
  return DefWindowProc(hwnd,message,wParam,lParam);
}

int GetTaskByWnd(HWND Hwnd)
{
  int i;
  for (i = 0;i < numTasks ; i++)
    if (tasks[i].hwnd == Hwnd)
	  return i;
  return -1;
}

int GetUnMovedNrTasks(void)
{
  int result = 0;
  for (int i = 0; i < numTasks; i++)
    if (!tasks[i].IsMoved)
	  result++;
  return result;
}

void CreateHints(HWND hWnd, char *txt, RECT *r)
{ /*from the original ls code (shortcuts)*/
  TOOLINFO ti;/*tool information*/
  if (ts.NoHints)
	return;
  ti.cbSize = sizeof(TOOLINFO);
  ti.uFlags = TTF_SUBCLASS;
  ti.hwnd = hWnd;
  ti.hinst = dll;
  ti.uId = 0;
  ti.lpszText = txt;
  ti.rect = *r;
  SendMessage(hints, TTM_ADDTOOL, 0, (LPARAM) (LPTOOLINFO) &ti);
  return;
}

void CreateTask(int i, HWND handle)
{
  RECT r;
  HWND hWnD = NULL;
  char titleTmp[MAXLEN], classTmp[MAXLEN];
  /*for always ontop -maduin*/
  hWnD = CreateWindowEx(WS_EX_TOOLWINDOW|WS_EX_TOPMOST, AppName, "", WS_POPUP, 0, 0, ts.Width, ts.Height, NULL, NULL, dll, NULL);
  if (!ts.AlwaysOnTop) SetParent(hWnD, deskTop);
  if (!hWnD) 
  {						   
    MessageBox(parent,"Error creating window","Tasks",MB_OK);
    return;
  }
  tasks[i].hwnd = hWnD;
  tasks[i].appHandle = handle;
  SetWindowLong(tasks[i].hwnd, GWL_USERDATA, magicDWord);
  {
	GetWindowText(handle, (char *)&titleTmp, MAXLEN);
	tasks[i].appTitle = new char[strlen(titleTmp) + 1];
	strcpy(tasks[i].appTitle, titleTmp);
  }
  {
	GetClassName(handle, (char *)&classTmp, MAXLEN);
	tasks[i].appClass = new char[strlen(classTmp) + 1];
	strcpy(tasks[i].appClass, classTmp);
  }
  tasks[i].IsActive = (GetForegroundWindow() == handle);
  tasks[i].IsIconic = IsIconic(handle);
  tasks[i].IsMoved  = false;
  tasks[i].IsCustomIcon = false;
  GetNewTaskPos(i);
  GetTasksIcon(i);
  GetTasksPix(i);
  SetWindowPos(tasks[i].hwnd, 0, tasks[i].x, tasks[i].y, ts.Width, ts.Height, SWP_NOZORDER|SWP_NOACTIVATE);
  if (ts.StartHidden)
	ShowWindow(tasks[i].hwnd, SW_HIDE);
  else
	ShowWindow(tasks[i].hwnd, SW_SHOWNORMAL|SW_SHOWNOACTIVATE);
  GetClientRect(tasks[i].hwnd, &r);
  CreateHints(tasks[i].hwnd, tasks[i].appTitle, &r);
  InvalidateRect(tasks[i].hwnd, &r, TRUE);
  return;
}

void DeleteTask(int i)
{
  RemoveHints(tasks[i].hwnd);
  if (tasks[i].hwnd)
    DestroyWindow(tasks[i].hwnd);
  if (tasks[i].pix)
    DeleteObject(tasks[i].pix);
  if (tasks[i].appClass)
    delete [] tasks[i].appClass;
  if (tasks[i].appTitle)
    delete [] tasks[i].appTitle;
  if (tasks[i].icon && tasks[i].IsCustomIcon)
	DestroyIcon(tasks[i].icon);
  tasks[i].appHandle = NULL;
  tasks[i].hwnd = NULL;
  tasks[i].pix = NULL;
  tasks[i].icon = NULL;  
  tasks[i].appTitle = NULL;
  tasks[i].appClass = NULL;
  return;
}

void GetNewTaskPos(int i)
{
  TaskPosition(i, (ts.AutoArrange ? GetUnMovedNrTasks(): numTasks));
  return;
}

void GetTasks(void)
{
  windowType * winList = NULL;
  char tmpTitle[MAXLEN], tmpClass[MAXLEN];
  int m, n, maxWin;
  BOOL refreshTask = false;
  prevTasks = numTasks;
  numTasks = 0;		  
  winList = (windowType *)SendMessage(parent, LM_WINLIST, 0, 0);
  maxWin = (int) SendMessage(parent, LM_WINLIST, 1, 0);
  for (n = 0; n < maxWin && winList[n].Handle; n++)
  {
	GetWindowText(winList[n].Handle, (char *)&tmpTitle, MAXLEN);
	GetClassName(winList[n].Handle, (char *)&tmpClass, MAXLEN);
	if ((winList[n].Handle != NULL && numTasks < MAXTASKS) && (!ts.OnlyMinimized || IsIconic(winList[n].Handle)) && IsAppWindow(winList[n].Handle) && !GetTasksIgnore(tmpTitle) && !GetTasksIgnore(tmpClass))
	{
	  if (winList[n].Handle != tasks[numTasks].appHandle)
	  { 
		DeleteTask(numTasks);
	    CreateTask(numTasks, winList[n].Handle);
//	    numTasks++;
	  }
	  if (tasks[numTasks].IsIconic != IsIconic(winList[n].Handle))
	  {
		tasks[numTasks].IsIconic = IsIconic(winList[n].Handle);
		if (tasks[numTasks].IsIconic && tasks[numTasks].IsActive)
		{/*gotta refresh the task if it had been active and highlighting is turned on*/
		  tasks[numTasks].IsActive = false;
		  if (ts.HighLightSelected)
			refreshTask = true;
		}
//		if (!tasks[numTasks].IsIconic && curTask == tasks[numTasks].appHandle)
//		  tasks[numTasks].IsActive = true;
//		if (!ts.StartHidden)
//		{/*Have to force a complete redraw here...*/
//		  ShowWindow(tasks[numTasks].hwnd, SW_HIDE);
//		  ShowWindow(tasks[numTasks].hwnd, SW_SHOWNORMAL);
//		}
//		numTasks++;
	  }
	  if (strcmp(tmpTitle, tasks[numTasks].appTitle) != 0)
	  {
		delete [] tasks[numTasks].appTitle;
		tasks[numTasks].appTitle = new char[strlen(tmpTitle) + 1];
		strcpy(tasks[numTasks].appTitle, tmpTitle);
		if (!ts.NoHints)
		{
		  TOOLINFO ti;
		  ti.cbSize = sizeof(TOOLINFO);
		  ti.hwnd = tasks[n].hwnd;
		  ti.uId = 0;
		  ti.lpszText = tmpTitle;
		  SendMessage(hints, TTM_UPDATETIPTEXT, 0, (LPARAM) (LPTOOLINFO) &ti);
		}
		if (/*!ts.StartHidden && */(ts.TitleAlways || (ts.TitleOnlyMinimized && tasks[numTasks].IsIconic)))
		{/*Have to force a complete redraw here...*/
		  refreshTask = true;
//		  ShowWindow(tasks[numTasks].hwnd, SW_HIDE);
//		  ShowWindow(tasks[numTasks].hwnd, SW_SHOWNORMAL);
		}
//		numTasks++;
	  }
	  if ((tasks[numTasks].IsActive != (curTask == tasks[numTasks].appHandle)) && ts.HighLightSelected && !tasks[numTasks].IsIconic) /*Strange instance where a minimized window can still be the 'ForegroundWindow'-active?!?*/
	  {
//		if (tasks[numTasks].IsIconic) /*wierd instance where a minimized window still has focus?!?*/
//		  tasks[numTasks].IsActive = false;
		if (curTask == tasks[numTasks].appHandle)
		  tasks[numTasks].IsActive = true;
		else
		  tasks[numTasks].IsActive = false;
		refreshTask = true;
//		if (!ts.StartHidden)
//		{/*Have to force a complete redraw here...*/
//		  ShowWindow(tasks[numTasks].hwnd, SW_HIDE);
//		  ShowWindow(tasks[numTasks].hwnd, SW_SHOWNORMAL);
//		}
//		numTasks++;
	  }
//	  else
//	    numTasks++;
	  if (tasks[numTasks].icon != GetIconFromWindow(winList[n].Handle, (ts.IconSize > 23) ? true: false) && !tasks[numTasks].IsCustomIcon)
	  {
		tasks[numTasks].icon = GetIconFromWindow(winList[n].Handle, (ts.IconSize > 23) ? true: false);
		refreshTask = true;
	  }
	  if (!ts.StartHidden && refreshTask)
	  {
		ShowWindow(tasks[numTasks].hwnd, SW_HIDE);
		ShowWindow(tasks[numTasks].hwnd, SW_SHOWNORMAL|SW_SHOWNOACTIVATE);
	  }
	  refreshTask = false;
	  numTasks++;
	}
  }
  for (m = numTasks; m < prevTasks; m++)
	DeleteTask(m);
  return;
}

void GetTasksIcon(int n)
{
  int i;
  for (i = 0; i < numIcon; i++)
  {
	if (strstr(tasks[n].appTitle, iconTag[i].match) || strstr(tasks[n].appClass, iconTag[i].match))
	{
	  tasks[n].icon = ExtractIcon(dll,iconTag[i].name,iconTag[i].icon);
	  if (tasks[n].icon)
	  {
		tasks[n].IsCustomIcon = true;
		return;
	  }
	}
  }
  tasks[n].icon = GetIconFromWindow(tasks[n].appHandle, (ts.IconSize > 23) ? true: false);
  return;
}

void GetTasksPix(int n)
{
  int i;
  for (i = 0; i < numPix; i++)
  {
	if (strstr(tasks[n].appTitle, pixTag[i].match) || strstr(tasks[n].appClass, pixTag[i].match))
	{
	  tasks[n].pix = LoadLSImage(pixTag[i].pix, NULL);
	  if (tasks[n].pix) tasks[n].icon = NULL;
	  return;
	}
  }
  return;
}

void PaintTask(int i)
{
  int posX, posY;
  BOOL activeTask = ts.HighLightSelected && tasks[i].IsActive;
  BOOL minTask = ts.TitleOnlyMinimized && tasks[i].IsIconic;
  COLORREF bgColor, darkColor, lightColor, titleBgColor, titleDarkColor, titleFontColor, titleLightColor;
  PAINTSTRUCT ps;
  BITMAP bmInfo;
//  HBITMAP backImage;
  HDC hdc = BeginPaint(tasks[i].hwnd,&ps);
  HDC src = CreateCompatibleDC(NULL);
  if (activeTask)
  {
//	backImage = ts.SelBackImage;
	bgColor = ts.SelBgColor;
	lightColor = ts.SelLightColor;
	darkColor = ts.SelDarkColor;
	titleBgColor = ts.TitleSelBgColor;
	titleDarkColor = ts.TitleSelDarkColor;
	titleFontColor = ts.TitleSelFontColor;
	titleLightColor = ts.TitleSelLightColor;
  }
  else
  {
//	backImage = ts.BackImage;
	bgColor = ts.BgColor;
	lightColor = ts.LightColor;
	darkColor = ts.DarkColor;
	titleBgColor = ts.TitleBgColor;
	titleDarkColor = ts.TitleDarkColor;
	titleFontColor = ts.TitleFontColor;
	titleLightColor = ts.TitleLightColor;
  }

  {/*paint the background unless transparent*/
	RECT r;
	GetClientRect(tasks[i].hwnd, &r);
	if (bgColor != 0x00FF00FF)
	{
	  HBRUSH backBrush;
	  backBrush = CreateSolidBrush(bgColor);
	  FillRect(hdc, &r, backBrush);
	  DeleteObject(backBrush);
	}
  }
  if ((ts.BackImage || (ts.SelBackImage && activeTask)) && !(selBgImgNone && activeTask))
  {
//	GetObject(backImage, sizeof(bmInfo), &bmInfo);
//	SelectObject(src, backImage);
	if (ts.SelBackImage && activeTask)
	{
	  GetObject(ts.SelBackImage, sizeof(bmInfo), &bmInfo);
	  SelectObject(src, ts.SelBackImage);
	}
	else
	{
	  GetObject(ts.BackImage, sizeof(bmInfo), &bmInfo);
	  SelectObject(src, ts.BackImage);
	}
	posX = ts.Width/2 - bmInfo.bmWidth/2;   /*Center in Tile*/
	posY = ts.Height/2 - bmInfo.bmHeight/2;
    TransparentBltLS(hdc, posX, posY, bmInfo.bmWidth, bmInfo.bmHeight, src, 0, 0, RGB(255, 0, 255));
  }
  if (tasks[i].icon)
  {
	posX = ts.IconX;
	posY = ts.IconY;
	if (ts.TitleAlways || minTask)
	{
	  posX = ts.TitleIconX;
	  posY = ts.TitleIconY;
	}
	DrawIconEx(hdc, posX, posY, tasks[i].icon, ts.IconSize, ts.IconSize, 0, NULL, DI_NORMAL);
  }
  if (tasks[i].pix)
  {
    GetObject(tasks[i].pix, sizeof(bmInfo), &bmInfo);
	/*Could possibly add code for resizing the pix if larger then available space, with StretchBLT() or something. I don't think its needed tho. -jugg*/
	posX = ts.PixX;
	posY = ts.PixY;
	if (ts.TitleAlways || minTask)
	{
	  posX = ts.TitlePixX;
	  posY = ts.TitlePixY;
	}
    SelectObject(src, tasks[i].pix);
	TransparentBltLS(hdc, posX, posY, bmInfo.bmWidth, bmInfo.bmHeight, src, 0, 0, RGB(255,0,255));
  }
  if (ts.TitleAlways || minTask)
  {
	RECT r;
	GetClientRect(tasks[i].hwnd, &r);
	  r.top = ts.TitleY;
	  r.left = ts.TitleX;
	  r.bottom = ts.TitleHeight + ts.TitleY;
	  r.right = ts.TitleWidth + ts.TitleX;
	if (titleBgColor != 0x00FF00FF)
	{
	  HBRUSH backBrush;
	  backBrush = CreateSolidBrush(titleBgColor);
	  FillRect(hdc, &r, backBrush);
	  DeleteObject(backBrush);
	}
	if (titleLightColor != 0x00FF00FF)
	{
	  HPEN pen, oldpen;
	  /*Light Color Bevel (left/top)*/
	  pen = CreatePen(PS_SOLID, 1, titleLightColor);
	  oldpen = (HPEN) SelectObject(hdc, pen);
	  MoveToEx(hdc, r.left, r.bottom - 1, NULL);
	  LineTo(hdc, r.left, r.top);
	  LineTo(hdc, r.right - 1, r.top);
	  SelectObject(hdc, oldpen);
	  DeleteObject(pen);
	}
	if (titleDarkColor != 0x00FF00FF)
	{
	  HPEN pen, oldpen;
	  /*Dark Color Bevel (bottom/right)*/
	  pen = CreatePen(PS_SOLID, 1, titleDarkColor);
	  oldpen = (HPEN) SelectObject(hdc, pen);
	  MoveToEx(hdc, r.right - 1, r.top, NULL);
	  LineTo(hdc, r.right - 1, r.bottom - 1);
	  LineTo(hdc, r.left, r.bottom - 1);
	  SelectObject(hdc, oldpen);
	  DeleteObject(pen);
	}
    HFONT titleFont, oldFont;
    int len = strlen(tasks[i].appTitle);
	r.top += 1;
	r.bottom -= 1;
	r.left += 2;
	r.right -= 2;
    titleFont = CreateFont(ts.TitleFontSize, 0, 0, 0, 0, FALSE, FALSE, FALSE, ANSI_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH | FF_DONTCARE, ts.TitleFont);
    oldFont = (HFONT) SelectObject(hdc, titleFont);
    SetBkMode(hdc, TRANSPARENT);
    SetTextColor(hdc, titleFontColor);
	DrawText(hdc, tasks[i].appTitle, len, &r, DT_VCENTER|DT_SINGLELINE|DT_END_ELLIPSIS);
    SelectObject(hdc, oldFont);
    DeleteObject(titleFont);
  }
  {
	RECT r;
	GetClientRect(tasks[i].hwnd, &r);
	if (lightColor != 0x00FF00FF)
	{
	  HPEN pen, oldpen;
	  /*Light Color Bevel (left/top)*/
	  pen = CreatePen(PS_SOLID, 1, lightColor);
	  oldpen = (HPEN) SelectObject(hdc, pen);
	  MoveToEx(hdc, r.left, r.bottom - 1, NULL);
	  LineTo(hdc, r.left, r.top);
	  LineTo(hdc, r.right - 1, r.top);
	  SelectObject(hdc, oldpen);
	  DeleteObject(pen);
	}
	if (darkColor != 0x00FF00FF)
	{
	  HPEN pen, oldpen;
	  /*Dark Color Bevel (bottom/right)*/
	  pen = CreatePen(PS_SOLID, 1, darkColor);
	  oldpen = (HPEN) SelectObject(hdc, pen);
	  MoveToEx(hdc, r.right - 1, r.top, NULL);
	  LineTo(hdc, r.right - 1, r.bottom - 1);
	  LineTo(hdc, r.left, r.bottom - 1);
	  SelectObject(hdc, oldpen);
	  DeleteObject(pen);
	}
  }
  DeleteDC(src);
  EndPaint(tasks[i].hwnd,&ps);
//  if (backImage)
//	DeleteObject(backImage);
  return;
}

void ReadTasksIcon(void)
{
  FILE * f;
  unsigned int i;
  char * token[4];
  char buffer[1024];

  numIcon = 0;
  f = LCOpen(NULL);
  if (!f) 
	return;
  for (i = 0; i < 4; i++)
    token[i] = new char[MAXLEN];
  while (LCReadNextConfig (f, "*TasksIcon", buffer, sizeof (buffer)))
  {
	 int count;
	 count = LCTokenize (buffer, token, 4, NULL);
	 if (count >= 4)
	 {
	   iconTag[numIcon].match = new char[strlen(token[1])+1];
	   iconTag[numIcon].name = new char[strlen(token[2])+1];
	   strcpy(iconTag[numIcon].match, token[1]);
	   strcpy(iconTag[numIcon].name, token[2]);
	   iconTag[numIcon].icon = atoi(token[3]);
	   numIcon++;
	 }
  }
  LCClose(f);
  for (i = 0; i < 4; i++)
    delete [] token[i];
  return;
}

void ReadTasksIgnore(void)
{
  FILE * f;
  unsigned int i;
  char * token[2];
  char buffer[1024];

  numIgnore = 0;
  f = LCOpen(NULL);
  if (!f) 
	return;
  for (i = 0; i < 2; i++)
    token[i] = new char[MAXLEN];
  while (LCReadNextConfig (f, "*TasksIgnore", buffer, sizeof (buffer)))
  {
	 int count;
	 count = LCTokenize (buffer, token, 2, NULL);
	 if (count >= 2)
	 {
	   ignoreTag[numIgnore].match = new char[strlen(token[1])+1];
	   strcpy(ignoreTag[numIgnore].match, token[1]);
	   numIgnore++;
	 }
  }
  LCClose(f);
  for (i = 0; i < 2; i++)
    delete [] token[i];
  return;
}

void ReadTasksPix(void)
{
  FILE * f;
  unsigned int i;
  char * token[3];
  char buffer[1024];

  numPix = 0;
  f = LCOpen(NULL);
  if (!f) 
	return;
  for (i = 0; i < 3; i++)
    token[i] = new char[MAXLEN];
  while (LCReadNextConfig (f, "*TasksPix", buffer, sizeof (buffer)))
  {
	 int count;
	 count = LCTokenize (buffer, token, 3, NULL);
	 if (count >= 3)
	 {
	   pixTag[numPix].match = new char[strlen(token[1])+1];
	   pixTag[numPix].pix = new char[strlen(token[2])+1];
	   strcpy(pixTag[numPix].match, token[1]);
	   strcpy(pixTag[numPix].pix, token[2]);
	   numPix++;
	 }
  }
  LCClose(f);
  for (i = 0; i < 3; i++)
    delete [] token[i];
  return;
}

void ReArrangeTasks(BOOL all)
{
  int j = 0;
  for (int i = 0; i < numTasks; i++)
  {
	if (all || !tasks[i].IsMoved)
	{
	  TaskPosition(i, j);
      SetWindowPos(tasks[i].hwnd, 0, tasks[i].x, tasks[i].y, ts.Width, ts.Height, SWP_NOZORDER|SWP_NOACTIVATE);
	  j++;
	  tasks[i].IsMoved = false;
	  if (!ts.StartHidden)
	  {/*Have to force a complete redraw here...*/
		ShowWindow(tasks[i].hwnd, SW_HIDE);
		ShowWindow(tasks[i].hwnd, SW_SHOWNORMAL);
	  }
	}
  }
  all = false;
  return;
}

void RemoveHints(HWND hWnd)
{/*from the original ls code (shortcuts)*/
  TOOLINFO ti;/*tool information*/
  RECT r={0,0,0,0};
  ti.cbSize = sizeof(TOOLINFO);
  ti.uFlags = 0;
  ti.hwnd = hWnd;
  ti.hinst = dll;
  ti.uId = 0;
  ti.lpszText = NULL;
  ti.rect = r;
  SendMessage(hints, TTM_DELTOOL, 0, (LPARAM) (LPTOOLINFO) &ti);
  return;
}

void TaskPosition(int i, int tmpNumTasks)
{
  int wrapLine = 0, resetLine = 0, checkLine = 0;
  wrapLine = tmpNumTasks / ts.WrapCount;
  checkLine = tmpNumTasks - ts.WrapCount * wrapLine;
  if (checkLine == 1)
	resetLine = 1;
  else
	resetLine = tmpNumTasks - ts.WrapCount * wrapLine;
  if (ts.Direction == 1) /*up*/
  {
	if (taskWrap)
	{
	  if (ts.Y >= 0)
	    tasks[i].y = ts.Y - ts.Height - resetLine * (ts.Height + ts.PaddingY);
	  else /*now accepts negative values*/
	    tasks[i].y = screenHeight + ts.Y - ts.Height - resetLine * (ts.Height + ts.PaddingY);
	  if (ts.WrapDirection == 4) /*left*/
	  {
		if (ts.X >= 0)
		  tasks[i].x = ts.X - ts.Width - wrapLine * (ts.Width + ts.PaddingX);
		else /*now accepts negative values*/
		  tasks[i].x = screenWidth + ts.X - ts.Width - wrapLine * (ts.Width + ts.PaddingX);
	  }
	  else /*defaults to right, even if user inputs invalid up or down setting*/
	  {
	    if (ts.X >= 0)
	      tasks[i].x = ts.X + wrapLine * (ts.Width + ts.PaddingX);
	    else /*now accepts negative values*/
	      tasks[i].x = screenWidth + ts.X + wrapLine * (ts.Width + ts.PaddingX);
	  }
	}
	else
	{
	  if (ts.X >= 0)
	    tasks[i].x = ts.X;
	  else /*now accepts negative values*/
	    tasks[i].x = screenWidth + ts.X;
	  if (ts.Y >= 0)
	    tasks[i].y = ts.Y - ts.Height - tmpNumTasks * (ts.Height + ts.PaddingY);
	  else /*now accepts negative values*/
	    tasks[i].y = screenHeight + ts.Y - ts.Height - tmpNumTasks * (ts.Height + ts.PaddingY);
	}
  }
  else if (ts.Direction == 2) /*right*/
  {
    if (taskWrap)
	{
	  if (ts.X >= 0)
		tasks[i].x = ts.X + resetLine * (ts.Width + ts.PaddingX);
	  else /*now accepts negative values*/
	    tasks[i].x = screenWidth + ts.X + resetLine * (ts.Width + ts.PaddingX);
	  if (ts.WrapDirection == 1) /*up*/
	  {
	    if (ts.Y >= 0)
	      tasks[i].y = ts.Y - ts.Height - wrapLine * (ts.Height + ts.PaddingY);
	    else /*now accepts negative values*/
		  tasks[i].y = screenHeight + ts.Y - ts.Height - wrapLine * (ts.Height + ts.PaddingY);
	  }
	  else /*defaults to down, even if user inputs invalid left or right setting*/
	  {
	    if (ts.Y >= 0)
	      tasks[i].y = ts.Y + wrapLine * (ts.Height + ts.PaddingY);
	    else /*now accepts negative values*/
	      tasks[i].y = screenHeight + ts.Y + wrapLine * (ts.Height + ts.PaddingY);
	  }
	}
	else
	{
	  if (ts.X >= 0)
		tasks[i].x = ts.X + tmpNumTasks * (ts.Width + ts.PaddingX);
	  else /*now accepts negative values*/
	    tasks[i].x = screenWidth + ts.X + tmpNumTasks * (ts.Width + ts.PaddingX);
	  if (ts.Y >= 0)
		tasks[i].y = ts.Y;
	  else /*now accepts negative values*/
		tasks[i].y = screenHeight + ts.Y;
	}
  }
  else if (ts.Direction == 3) /*down*/
  {
    if (taskWrap)
	{
	  if (ts.Y >= 0)
	    tasks[i].y = ts.Y + resetLine * (ts.Height + ts.PaddingY);
	  else /*now accepts negative values*/
	    tasks[i].y = screenHeight + ts.Y + resetLine * (ts.Height + ts.PaddingY);
	  if (ts.WrapDirection == 4) /*left*/
	  {
		if (ts.X >= 0)
		  tasks[i].x = ts.X - ts.Width - wrapLine * (ts.Width + ts.PaddingX);
		else /*now accepts negative values*/
		  tasks[i].x = screenWidth + ts.X - ts.Width - wrapLine * (ts.Width + ts.PaddingX);
	  }
	  else /*defaults to right, even if user inputs invalid up or down setting*/
	  {
	    if (ts.X >= 0)
	      tasks[i].x = ts.X + wrapLine * (ts.Width + ts.PaddingX);
	    else /*now accepts negative values*/
	      tasks[i].x = screenWidth + ts.X + wrapLine * (ts.Width + ts.PaddingX);
	  }
	}
	else
	{
	  if (ts.X >= 0)
	    tasks[i].x = ts.X;
	  else /*now accepts negative values*/
	    tasks[i].x = screenWidth + ts.X;
	  if (ts.Y >= 0)
	    tasks[i].y = ts.Y + tmpNumTasks * (ts.Height + ts.PaddingY);
	  else /*now accepts negative values*/
	    tasks[i].y = screenHeight + ts.Y + tmpNumTasks * (ts.Height + ts.PaddingY);
	}
  }
  else if (ts.Direction == 4) /*left*/
  {
	if (taskWrap)
	{
	  if (ts.X >= 0)
	    tasks[i].x = ts.X - ts.Width - resetLine * (ts.Width + ts.PaddingX);
	  else /*now accepts negative values*/
	    tasks[i].x = screenWidth + ts.X - ts.Width - resetLine * (ts.Width + ts.PaddingX);
	  if (ts.WrapDirection == 1) /*up*/
	  {
	    if (ts.Y >= 0)
	      tasks[i].y = ts.Y - ts.Height - wrapLine * (ts.Height + ts.PaddingY);
	    else /*now accepts negative values*/
		  tasks[i].y = screenHeight + ts.Y - ts.Height - wrapLine * (ts.Height + ts.PaddingY);
	  }
	  else /*defaults to down, even if user inputs invalid left or right setting*/
	  {
	    if (ts.Y >= 0)
	      tasks[i].y = ts.Y + wrapLine * (ts.Height + ts.PaddingY);
	    else /*now accepts negative values*/
	      tasks[i].y = screenHeight + ts.Y + wrapLine * (ts.Height + ts.PaddingY);
	  }
	}
	else
	{
	  if (ts.X >= 0)
	    tasks[i].x = ts.X - ts.Width - tmpNumTasks * (ts.Width + ts.PaddingX);
	  else /*now accepts negative values*/
	    tasks[i].x = screenWidth + ts.X - ts.Width - tmpNumTasks * (ts.Width + ts.PaddingX);
      if (ts.Y >= 0)
        tasks[i].y = ts.Y;
      else /*now accepts negative values*/
        tasks[i].y = screenHeight + ts.Y;
	}
  }
  return;
}

/*From here down are the !bang commands*/
void BottomTasks(void)
{
  for (int i = 0; i < numTasks; i++)
  {
	SetParent(tasks[i].hwnd, deskTop);
	SetWindowPos(tasks[i].hwnd,HWND_NOTOPMOST,0,0,0,0,SWP_NOMOVE|SWP_NOSIZE|SWP_NOZORDER|SWP_NOACTIVATE);
  }
  ts.AlwaysOnTop = false;
  return;
}

void GatherTasks(void)
{
  ReArrangeTasks(true);
  return;
}

void HideTasks(void)
{
  for (int i = 0; i < numTasks; i++)
  {
	ShowWindow(tasks[i].hwnd, SW_HIDE);
  }
  ts.StartHidden = true;
  return;
}

void ShowTasks(void)
{
  for (int i = 0; i < numTasks; i++)
  {
	ShowWindow(tasks[i].hwnd, SW_SHOWNORMAL);
  }
  ts.StartHidden = false;
  return;
}

void ToggleTasks(void)
{
  for (int i = 0; i < numTasks; i++)
  {
	if (ts.StartHidden)
	  ShowWindow(tasks[i].hwnd, SW_SHOWNORMAL);
	else
	  ShowWindow(tasks[i].hwnd, SW_HIDE);
  }
  ts.StartHidden = !ts.StartHidden;
  return;
}

void TopTasks(void)
{
  for (int i = 0; i < numTasks; i++)
  {
	SetParent(tasks[i].hwnd, NULL);
	SetWindowPos(tasks[i].hwnd,HWND_TOPMOST,0,0,0,0,SWP_NOMOVE|SWP_NOSIZE|SWP_NOZORDER|SWP_NOACTIVATE);
  }
  ts.AlwaysOnTop = true;
  return;
}